import json
import logging
from dataclasses import dataclass
from datetime import datetime, timedelta
from enum import Enum
from typing import Optional

import requests
from bs4 import BeautifulSoup


DEFAULT_TIME = datetime.now() + timedelta(days=1000)


class VideoSite(Enum):
    NICO_NICO = "niconico"
    BILIBILI = "bilibili"
    YOUTUBE = "YouTube"


@dataclass
class Video:
    site: VideoSite = None
    identifier: str = None
    url: str = None
    views: int = 0
    uploaded: datetime = DEFAULT_TIME
    thumb_url: str = None
    canonical = True

    def __str__(self) -> str:
        return f"VideoSite: {self.site}\n" \
               f"Id: {self.identifier}\n" \
               f"Views: {self.views}\n" \
               f"Uploaded: {self.uploaded}\n" \
               f"Thumb: {self.thumb_url}\n\n"


table = 'fZodR9XQDSUm21yCkr6zBqiveYah8bt4xsWpHnJE7jL5VG3guMTKNPAwcF'
tr = {}
for index in range(58):
    tr[table[index]] = index
s = [11, 10, 3, 8, 4, 6]
xor = 177451812
add = 8728348608


def av_to_bv(av: str) -> str:
    x = int(av[2:])
    x = (x ^ xor) + add
    r = list('BV1  4 1 7  ')
    for i in range(6):
        r[s[i]] = table[x // 58 ** i % 58]
    return ''.join(r)


def str_to_date(date: str) -> Optional[datetime]:
    if 'T' in date:
        date = date[:date.find('T')]
    date = date.split("-")
    if len(date) != 3:
        return None
    year = int(date[0])
    month = int(date[1])
    day = int(date[2])
    return datetime(year=year, month=month, day=day)


def get_nc_info(vid: str) -> Video:
    if vid.find("nicovideo") != -1:
        vid = vid[vid.rfind("/") + 1:]
    url = f"https://www.nicovideo.jp/watch/{vid}"
    result = requests.get(url).text
    soup = BeautifulSoup(result, "html.parser")
    date = DEFAULT_TIME
    views = 0
    for script in soup.find_all('script'):
        t: str = script.get_text()
        index_start = t.find("uploadDate")
        if index_start != -1:
            index_start += len("uploadDate") + 3
            date = t[index_start:index_start + 10]
            date = str_to_date(date)
        index_start = t.find("userInteractionCount")
        if index_start != -1:
            index_start += len("userInteractionCount") + 2
            index_end = t.find("}", index_start)
            views = int(t[index_start:index_end])
    thumb = soup.find("meta", {"name": "thumbnail"})['content']
    return Video(VideoSite.NICO_NICO, vid, url, views, date, thumb)


def get_bb_info(vid: str) -> Video:
    if vid.find("bilibili") != -1:
        vid = vid[vid.rfind("/") + 1:]
        if vid.find("?") != -1:
            vid = vid[:vid.find("?")]
    if "av" in vid:
        vid = av_to_bv(vid)
    url = f"https://api.bilibili.com/x/web-interface/view?bvid={vid}"
    response = json.loads(requests.get(url).text)
    epoch_time = int(response['data']['pubdate'])
    date = datetime.fromtimestamp(epoch_time)
    # remove extra information to be in sync with YT and Nico
    date = datetime(year=date.year, month=date.month, day=date.day)
    pic = response['data']['pic']
    views = response['data']['stat']['view']
    return Video(VideoSite.BILIBILI, vid, url, views, date, pic)


def get_yt_info(vid: str) -> Video:
    if vid.find("youtube.") != -1:
        vid = vid[vid.find("=") + 1:]
    elif vid.find("youtu.be") != -1:
        vid = vid[vid.rfind("/") + 1:]
    url = 'https://www.youtube.com/watch?v=' + vid
    text = requests.get(url).text
    soup = BeautifulSoup(text, "html.parser")
    interaction = soup.select_one('meta[itemprop="interactionCount"][content]')
    views = int(interaction['content'])
    date = str_to_date(soup.select_one('meta[itemprop="datePublished"][content]')['content'])
    return Video(VideoSite.YOUTUBE, vid, url, views, date,
                 thumb_url="https://img.youtube.com/vi/{}/0.jpg".format(vid))


info_func = {
    VideoSite.NICO_NICO: get_nc_info,
    VideoSite.BILIBILI: get_bb_info,
    VideoSite.YOUTUBE: get_yt_info
}


def video_from_site(site: VideoSite, identifier: str, canonical: bool = True) -> Optional[Video]:
    try:
        v = info_func[site](identifier)
    except Exception as e:
        logging.info(e)
        return None
    if not v:
        return None
    v.canonical = canonical
    return v



from mgp_common.japanese import is_japanese


def is_empty(s: str) -> bool:
    return s is None or s == '' or s.isspace()


def auto_lj(s: str) -> str:
    if is_japanese(s):
        return "{{lj|" + s + "}}"
    return s



import re

hiragana_pattern = re.compile("[\u3041-\u3096]")
katakana_pattern = re.compile("[\u30A0-\u30FF]")
kanji_pattern = re.compile("[\u3400-\u4DB5\u4E00-\u9FCB\uF900-\uFA6A]")
punctuation_pattern = re.compile("[\uFF5F-\uFF9F]")


def is_hiragana(c: str) -> bool:
    return not not hiragana_pattern.fullmatch(c)


def is_katakana(c: str) -> bool:
    # FIXME: is this the right place to deal with the dot?
    return katakana_pattern.fullmatch(c) is not None and c != '・'


def is_kana(c: str) -> bool:
    return is_hiragana(c) or is_katakana(c)


def is_kanji(c: str) -> bool:
    return c == "々" or not not kanji_pattern.fullmatch(c)


def is_punctuation_or_half_width_katakana(c: str) -> bool:
    return not not punctuation_pattern.fullmatch(c)


def is_japanese(s: str) -> bool:
    return any(is_kana(c) or is_kanji(c) or is_punctuation_or_half_width_katakana(c) for c in s)



from pathlib import Path
from typing import Union

cache_path = Path("cache")


def set_cache_path(path: Union[str, Path]):
    global cache_path
    cache_path = Path(path)


def get_cache_path() -> Path:
    return cache_path



import json
import pickle
import re
from dataclasses import dataclass, field
from datetime import datetime, timedelta
from typing import List, Dict, Optional

import requests

import mgp_common.video
from mgp_common.config import get_cache_path
from mgp_common.string_utils import is_empty
from mgp_common.video import Video, VideoSite, video_from_site


@dataclass
class Song:
    name_ja: str
    name_chs: str = None
    name_other: List[str] = field(default_factory=list)
    original: bool = True
    publish_date: datetime = datetime.now() + timedelta(days=1000)
    videos: List[Video] = field(default_factory=list)
    albums: List[str] = field(default_factory=list)
    creators: Dict[str, List[str]] = field(default_factory=dict)


def parse_videos(videos: list, load_videos: bool = True) -> List[Video]:
    service_to_site: dict = {
        'NicoNicoDouga': VideoSite.NICO_NICO,
        'Youtube': VideoSite.YOUTUBE
    }
    result = []
    for v in videos:
        service = v['service']
        if v['pvType'] == 'Original' and service in service_to_site.keys():
            url = v['url']
            if load_videos:
                video = video_from_site(service_to_site.pop(service), url)
            else:
                video = Video(site=service_to_site.pop(service), url=url)
            if video:
                result.append(video)
    return result


def parse_albums(albums: list) -> List[str]:
    return [a['defaultName'] for a in albums]


vocaloid_names = {
    '初音ミク': "初音未来",
    '鏡音リン': "镜音铃",
    '鏡音レン': "镜音连",
    '巡音ルカ': "巡音流歌",
    'カイト': "KAITO",
    'メイコ': "MEIKO",
    '音街ウナ': "音街鳗",
    '歌愛ユキ': "歌爱雪",
    '結月ゆかり': "结月缘",
    '神威がくぽ': "神威乐步",
    'ブイフラワ': "v flower",
    'イア': "IA",
    'マユ': "MAYU",
    'GUMI': "GUMI",
    'ふかせ': "Fukase",
    'ギャラ子': 'Galaco',
    '心華': "心华",
    "紲星あかり": "绁星灯",
    '重音テト': "重音Teto",
    '鳴花ヒメ': '鸣花姬',
    '鳴花ミコト': '鸣花尊',
    '시유': 'SeeU',
    'Eleanor Forte': '爱莲娜·芙缇',
    'KAFU': '可不',
    'SeKai': '星界'
}


def name_shorten(name: str) -> str:
    for n in vocaloid_names.keys():
        if n in name:
            return n
    return name


def split_names(regex: str = "[・/，, ]+", text: str = "") -> List[str]:
    return re.split(regex, text)


def parse_creators(artists: Optional[list], artist_string: Optional[str]) -> Dict[str, List[str]]:
    if artists is None:
        artists = []
    if is_empty(artist_string) or len(split_names(text=artist_string)) != 2:
        artist_string = "feat."
    mapping: Dict[str, List[str]] = {}
    for artist in artists:
        if 'artist' in artist:
            name = artist['artist']['name']
            if artist['artist']['artistType'] == 'Vocaloid':
                # shorten names like 初音ミク V4X
                name = name_shorten(name)
        else:
            name = artist['name']
        roles = artist['roles']
        if roles == 'Default':
            roles = artist['categories']
        if roles == 'Other':
            continue
        roles = split_names(text=roles)
        name = name.strip()
        for role in roles:
            role = role.strip()
            if role in mapping:
                mapping[role].append(name)
            else:
                mapping[role] = [name]
    if "Vocalist" not in mapping:
        names = split_names(text=artist_string.split("feat.")[1])
        mapping['Vocalist'] = [name_shorten(n.strip()) for n in names if not is_empty(n)]
    if "Producer" not in mapping:
        names = split_names(text=artist_string.split("feat.")[0])
        mapping['Producer'] = [n.strip() for n in names if not is_empty(n)]
    return mapping


def get_song_by_id(song_id: str, load_videos: bool = True) -> Song:
    url = f"https://vocadb.net/api/songs/{song_id}/details"
    response = json.loads(requests.get(url).text)
    name_ja = response['song']['defaultName']
    additional_names = [n for n in re.split(", *", response['additionalNames'])
                        if not is_empty(n)]
    publish_date = mgp_common.video.str_to_date(response['song'].get('publishDate', ""))
    if publish_date is None:
        publish_date = datetime.now() + timedelta(days=1000)
    original = response['song']['songType'] == 'Original'
    videos = parse_videos(response['pvs'], load_videos)
    for v in videos:
        if v.uploaded is not None and v.uploaded < publish_date:
            publish_date = v.uploaded
    albums = parse_albums(response['albums'])
    creators = parse_creators(response['artists'], response['artistString'])
    return Song(name_ja=name_ja, name_other=additional_names,
                original=original, publish_date=publish_date,
                videos=videos, albums=albums, creators=creators)


def get_producer_songs(producer_id: str, load_videos: bool = False) -> List[Song]:
    path = get_cache_path().joinpath("producer_songs_" + producer_id + ".pickle")
    if not path.exists():
        start = 0
        max_results = 50
        result = []
        while True:
            url = "https://vocadb.net/api/songs"
            response = requests.get(url, params={
                'start': start,
                'query': "",
                'maxResults': max_results,
                'sort': 'PublishDate',
                'artistId[]': producer_id,
                'artistParticipationStatus': 'Everything'
            })
            items = response.json()['items']
            for item in items:
                result.append(get_song_by_id(item['id'], load_videos))
            if len(items) < max_results:
                break
            start += max_results
        pickle.dump(result, open(path, "wb"))
    result = pickle.load(open(path, "rb"))
    return result


def get_producer_albums(producer_id: str, only_main: bool = True, only_original: bool = True) -> List[str]:
    start = 0
    max_results = 50
    result = []
    while True:
        url = "https://vocadb.net/api/albums"
        response = requests.get(url, params={
            'start': start,
            'query': "",
            'maxResults': max_results,
            'sort': 'ReleaseDate',
            'artistId[]': producer_id,
            'artistParticipationStatus': 'OnlyMainAlbums' if only_main else 'Everything',
            'discType': 'Album' if only_original else 'Unknown'
        }).json()
        for album in response['items']:
            result.append(album['defaultName'])
        if len(response['items']) < max_results:
            break
        start += max_results
    result.reverse()
    return result



from setuptools import setup, find_packages

setup(
    name='mgp-common',
    version="0.1.7",
    license='MIT',
    author="Peter Li",
    packages=find_packages('src'),
    package_dir={'': 'src'},
    url='https://github.com/syccxcc/MGP-common',
    install_requires=[
        'requests',
        'beautifulsoup4'
    ],

)
